// Author: Nicolas Bonneel
// Released: April 14th, 2016
// Credit the following article: 
//   @article{BPC16, 
//   author = {Bonneel, Nicolas and Peyr\'{e}, Gabriel and Cuturi, Marco}, 
//   title = {{Wasserstein Barycentric Coordinates: Histogram Regression Using Optimal Transport}}, 
//   journal = {ACM Transactions on Graphics (SIGGRAPH 2016)}, 
//   volume = {35}, 
//   number = {4}, 
//   year = {2016}, 
//   }
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#pragma once
#include <vector>
#include "lbfgs.h" 
#include "kernels.h"
#include "loss.h"
#include <iostream>
#include <stdio.h>
#include <cstring>


template<class KernelType> class Problem;                // Definition of a problem
template<class KernelType> class WassersteinBarycenter;  //  Forward Barycenter computation
template<class KernelType, class LossType> class WassersteinRegression;  // Barycentric coordinates computation

// helper to perform a regression without the above classes
// pdfs (in): a set of histograms.
// observed_pdf(in): the input histogram to find the barycentric coordinates among the pdfs
// mkernel(in): the Gibbs kernel. Typically, one of the gaussian convolutions implemented in kernel.h 
// nearest_pdf (out): the projection of observed_pdf within the set of histograms "pdfs"
// best_lambdas (out): the corresponding barycentric coordinates
// niter (in): number of iterations for the inverse problem (higher values => slower and more memory)
// compute_barycenter (in): tells whether we are interested in returning the barycenter in "nearest_pdf"
// npdfs (in): if npdfs>1, each histogram actually stores (consecutively) a set of histograms. The problem is solved jointly for all histograms. This is useful for BRDFs, color images etc.
template<class Kernel, class Loss>
void regress(std::vector<std::vector<double> > &pdfs, const std::vector<double> &observed_pdf, Kernel &mkernel, std::vector<double> &nearest_pdf, double* best_lambdas, int niter = 50, bool computer_barycenter = true, int npdfs = 1) {

	int K = pdfs.size();
	Problem<Kernel> problem(pdfs, observed_pdf, mkernel, &best_lambdas[0], npdfs);

	WassersteinRegression<Kernel, Loss > regression(&problem, niter, problem.kernel);	
	regression.regress(&best_lambdas[0]);

	if (computer_barycenter) {
		problem.set_weights(&best_lambdas[0]);
		WassersteinBarycenter<Kernel> baryResult(&problem, 100);

		nearest_pdf.resize(problem.N*npdfs);
		for (int i=0; i<npdfs; i++) {
			baryResult.compute_barycenter(i);
			std::memcpy(&nearest_pdf[i*problem.N], &baryResult.get_barycenter()[0], problem.N*sizeof(nearest_pdf[0]));
		}
	}
}



///////////////////////////////////// IMPLEMENTATION ///////////////////////////////////////////

template<typename T> T dot(const T* u, const T* v, int N) {
	double res = 0.;
	for (int i=0; i<N; i++) res += u[i]*v[i];
	return res;
}

template<class KernelType>
class Problem {
public:
	Problem(std::vector<std::vector<double> > &histograms, const std::vector<double> &observed_histogram, const KernelType &pkernel, double* weights = NULL, int npdfs = 1) {
		pdfs = histograms;
		K = pdfs.size();
		N = pdfs[0].size()/npdfs;
		observed_pdf = observed_histogram;
		kernel = pkernel;
		lambdas = new double[pdfs.size()];
		num_pdfs = npdfs;
		if (weights) { set_weights(weights); };

	}
	Problem(const Problem& p) {  // Kernel is passed by pointer ; lambdas are copied
		pdfs = p.pdfs;
		K = p.K;
		N = p.N;
		observed_pdf = p.observed_pdf;
		kernel = p.kernel;
		num_pdfs = p.num_pdfs;
		lambdas = new double[pdfs.size()];
		set_weights(p.lambdas);
	}
	~Problem() {
		delete[] lambdas;
	}
	void set_weights(const double* weights) {
		std::memcpy(lambdas, weights, K*sizeof(double));
	}

	std::vector<std::vector<double> > pdfs;
	std::vector<double> observed_pdf;
	KernelType kernel;
	double* lambdas;
	size_t N, K;
	int num_pdfs;
};



template<class KernelType>
class WassersteinBarycenter {
	template<class T, class S> 	friend  class WassersteinRegression;

public:
	WassersteinBarycenter(Problem<KernelType>* p, int n_bregman_iter) {
		K = p->pdfs.size();
		N = p->pdfs[0].size()/p->num_pdfs;
		problem = p;
		Niters = n_bregman_iter;
	}

	double get_plan(int id, int i, int j) {
		return a[id*N+i]*b[id*N+j]*problem->kernel(i, j);
	}

	// Bregman projections
	void compute_barycenter(int id = 0) {

		barycenter.resize(N);

		a.resize(K*N);
		b.resize(K*N);
		for (int i=0; i<K*N; i++){ b[i]=1.; }

		double* convolution = (double*) malloc_simd(K*N*sizeof(double), ALIGN);

		// Bregman Projections
		for (int iter=0; iter<Niters; iter++) {
			
			for (int i=0; i<K; i++) {

				problem->kernel.convolve(&b[i*N+0], &convolution[(size_t)i*N]);
				for (int j=0; j<N; j++) {
					a[i*N+j] = problem->pdfs[i][id*N + j] / convolution[i*N+j];
				}
				problem->kernel.convolveAdjoint(&a[i*N+0], &convolution[(size_t)i*N]);
			}

			geomMean(&convolution[0], &barycenter[0]);

			for (int i=0; i<K; i++) {				
				for (int j=0; j<N; j++) {
					b[i*N+j] = barycenter[j] / convolution[i*N+j];
				}

			}

		}
		
		free_simd(convolution);
	}

	// Accessors
	const std::vector<double>& get_barycenter() const { return barycenter; };


private:

	// geometric mean used for Bregman projections
	void geomMean(double* convolved_a, double* result) const {
		
		std::memset(&result[0], 0, N*sizeof(result[0]));		
		for (int i=0; i<K; i++) {			
			double l = problem->lambdas[i];
			for (int j=0; j<N; j++) {
				result[j] += safe_log(convolved_a[i*N+j])*l;
			}
		}
		for (int j=0; j<N; j++)
			result[j] = exp(result[j]);
	}
public:
	std::vector<double, aligned_allocator<double> > a, b;
	std::vector<double> barycenter;

	Problem<KernelType> *problem;
	size_t K, N;
	int Niters;
};







template<class KernelType, class LossType>
class WassersteinRegression {
public:
	WassersteinRegression(Problem<KernelType>* p, int n_bregman_iterations, const LossType &loss) {
		K = p->pdfs.size();
		N = p->pdfs[0].size()/p->num_pdfs;
		iteration = 0;
		bary_computation = new WassersteinBarycenter<KernelType>(p, n_bregman_iterations);
		problem = p;		
		this->loss = loss;
	}

	~WassersteinRegression() {
		delete bary_computation;
	}

	void regress(double* solution) {
		double residual;		
		lbfgs_parameter_t param;
		lbfgs_parameter_init(&param);
		param.linesearch = LBFGS_LINESEARCH_BACKTRACKING_WOLFE; 
		param.epsilon = 1E-10;
		param.max_iterations = 100;
		param.max_linesearch = 10;
		param.delta = 1E-10;
		iteration = 0;

		// change of variable y <- log(x)
		for (int i=0; i<K; i++) {
			solution[i] = log(solution[i]);			
		}
		int ret = lbfgs(K, solution, &residual, evaluate, progress, (void*)this, &param);		
		for (int i=0; i<K; i++) {
			solution[i] = exp(solution[i]);
		}
		
		// normalization
		double s = 0;
		for (int i=0; i<K; i++) {
			s += solution[i];
		}
		for (int i=0; i<K; i++) {
			solution[i] /= s;
		}
	}

	
private:
	
	LossType loss;	
	int iteration;
	
	// geometric mean used for Bregman projections
	void geomMean(const double* convolved_a, const double* b, double* lambdas, double *result) const {

		std::memset(&result[0], 0, N*sizeof(result[0]));

		for (int i=0; i<K; i++) {
			double l = lambdas[i];
			for (int j=0; j<N; j++) {
				result[j] += safe_log(convolved_a[i*N+j])*l;
			}
		}

		for (int j=0; j<N; j++)
			result[j] = exp(result[j]);
	}	

	void compute_gradient_sinkhorn(double* result, double* barycenter, int id) {
		const int n_iter_gradients = std::max(std::min(50, bary_computation->Niters), std::min(50+iteration*5, bary_computation->Niters));
		std::cout<<"nb iter gradient: "<<n_iter_gradients<<std::endl;

		std::vector<double, aligned_allocator<double> > a(N), u(N), convu(N);
		b.resize(K*N*(bary_computation->Niters+1));
		std::fill(b.begin(), b.begin()+K*N, 1.0);
		
		conv_b.resize(K*N*bary_computation->Niters);
		phi.resize(K*N*(bary_computation->Niters+1));
		g.resize(N);
		r.resize(K*N); std::memset(&r[0], 0, K*N*sizeof(r[0]));
		
		convu.resize(N);
	
		// Bregman Projections
		for (int iter=1; iter<=n_iter_gradients; iter++) {

			for (int i=0; i<K; i++) {
				size_t offset = (size_t)((iter-1)*K*N + i*N);
				problem->kernel.convolveAdjoint(&b[(iter-1)*K*N + i*N+0], &conv_b[offset]);
				for (int j=0; j<N; j++) {
					a[j] = problem->pdfs[i][id*N + j] / conv_b[offset+j];
				}
				problem->kernel.convolve(&a[0], &phi[iter*K*N + i*N]);
			}

			std::memset(&barycenter[0], 0, N*sizeof(barycenter[0]));
			for (int i=0; i<K; i++) {
				double l = problem->lambdas[i];
				for (int j=0; j<N; j++) {
					barycenter[j] += safe_log(phi[iter*K*N + i*N+j])*l;
				}
			}
			for (int j=0; j<N; j++)
				barycenter[j] = exp(barycenter[j]);
			
#pragma omp parallel for
			for (int i=0; i<K; i++) {
				for (int j=0; j<N; j++) {
					b[iter*K*N+i*N+j] = barycenter[j] / phi[iter*K*N + i*N+j];
				}
			}
		}

		int iter = n_iter_gradients; // we only compute the gradient for the last iteration
		loss.gradient(&barycenter[0], &problem->observed_pdf[id*N], N, &g[0]);

		for (int i=0; i<N; i++) {
			g[i] *= barycenter[i];
		}

		std::memset(result, 0, K*sizeof(result[0]));

		// applies B recursively
		for (int sub_iter = iter; sub_iter>=1; sub_iter--) {

			for (int i=0; i<K; i++) {
				double dotp = 0;
				for (int j=0; j<N; j++) {
					dotp+= log(phi[sub_iter*K*N + i*N+j]) * g[j];
					u[j] = (problem->lambdas[i]*g[j] - r[i*N+j])/phi[sub_iter*K*N + i*N+j];
 				}
				result[i] += dotp;

				if (sub_iter!=1) {
					problem->kernel.convolve(&u[0], &convu[0]);

					for (int j=0; j<N; j++) {
						convu[j] *= -problem->pdfs[i][id*N +j] / sqr(conv_b[(sub_iter-1)*K*N + i*N + j]);
					}
					problem->kernel.convolveAdjoint(&convu[0], &r[i*N]);
					for (int j=0; j<N; j++) {
						r[i*N+j] *= b[(sub_iter-1)*K*N + i*N + j];
					}
				}
			}
			if (sub_iter!=1) {
				std::memset(&g[0], 0, N*sizeof(g[0]));
				for (int i=0; i<K; i++) {
					for (int j=0; j<N; j++)
						g[j] +=r[i*N+j];
				}
			}
			
		}

	}



	// for l-bfgs
	static lbfgsfloatval_t evaluate(void *instance, const lbfgsfloatval_t *lambdas, lbfgsfloatval_t *gradient, const int n, const lbfgsfloatval_t step)
	 {

		WassersteinRegression<KernelType, LossType>* regression = (WassersteinRegression<KernelType, LossType>*)instance;
		Problem<KernelType>* prob = regression->problem;
		WassersteinBarycenter<KernelType>* bary = regression->bary_computation;

		// get the actual lambdas (without the log transform)
		double* lambdas2 = new double[prob->K];		
		for (int i=0; i<prob->K; i++) {
			lambdas2[i] = exp(lambdas[i]);
		}

		// normalize		
		double s = 0;
		for (int i=0; i<prob->K; i++) {
			s+= lambdas2[i];
		}
		for (int i=0; i<prob->K; i++) {
			lambdas2[i] /= s;
		}
		prob->set_weights(lambdas2);
		
		std::vector<double> barycenter(prob->N);
		std::vector<double> curGradient(prob->K);
		std::memset(gradient, 0, prob->K * sizeof(gradient[0]));
		lbfgsfloatval_t lossVal = 0;
		for (int id = 0; id<prob->num_pdfs; id++) {

			regression->compute_gradient_sinkhorn(&curGradient[0], &barycenter[0], id);

			double dotP = dot(&lambdas2[0], &curGradient[0], prob->K)/s;
			for (int i=0; i<prob->K; i++) {
				gradient[i] += exp(lambdas[i]) * (curGradient[i] - dotP)/s;  // Jacobian of the change of variable
				std::cout<<gradient[i]<<std::endl;
			}
			lossVal += regression->loss.loss(&barycenter[0], &prob->observed_pdf[id*prob->N], prob->N);
		}
		
	
		delete[] lambdas2;
		
		return lossVal;
	}


	static int progress(void *instance, const lbfgsfloatval_t *lambdas, const lbfgsfloatval_t *gradient, const lbfgsfloatval_t fx, const lbfgsfloatval_t xnorm, const lbfgsfloatval_t gnorm, const lbfgsfloatval_t step, int n, int k, int ls)
	{
		WassersteinRegression<KernelType, LossType>* regression = (WassersteinRegression<KernelType, LossType>*)instance;
		regression->iteration++;
		printf("Iteration %d:\n", k);

		for (int i=0; i<n; i++) {
			printf("%f\t", exp(lambdas[i]));
		}

		printf("\nValue = %f\n", fx);
		return 0;
	}

	
	WassersteinBarycenter<KernelType>* bary_computation;
	Problem<KernelType>* problem;
	size_t K, N;	
	std::vector<double, aligned_allocator<double> > b, conv_b, phi, g, r;
};






