// Author: Nicolas Bonneel
// Released: April 14th, 2016
// Credit the following article: 
//   @article{BPC16, 
//   author = {Bonneel, Nicolas and Peyr\'{e}, Gabriel and Cuturi, Marco}, 
//   title = {{Wasserstein Barycentric Coordinates: Histogram Regression Using Optimal Transport}}, 
//   journal = {ACM Transactions on Graphics (SIGGRAPH 2016)}, 
//   volume = {35}, 
//   number = {4}, 
//   year = {2016}, 
//   }
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#pragma once
#include <algorithm>
#include <omp.h>
#include <math.h>
#include <assert.h>
#include <cstring>
#include <cmath>

#define SSE
#define AVX_SUPPORT
#include "sse_helpers.h"

// Epsilon for log(std::max(EPSILON, x)) and std::max(EPSILON, Gaussian(x))
#define EPSILON 1E-250

// Generic
class MatrixKernel;  // General matrix, with values actually stored. For non-symmetric distances
class SymmetricMatrixKernel; // Symmetric matrix, with values actually stored. For symmetric distances

// 1D Gaussian convolution
class Gaussian1DKernel;

// 2D Gaussian convolution
class Gaussian2DKernel;   // 2D Gaussian Kernel, for 2-Wasserstein. 
class GaussianSSE2DKernel; // same, vectorized with SSE/AVX  (4x speedup ; 1E-14% error)
class TruncatedGaussian2DKernel; // truncated Gaussian ; inaccurate, even for largely truncated gaussians

// 3D Gaussian convolution (same suffixes)
class Gaussian3DKernel;
class TruncatedGaussian3DKernel;

// 4D Gaussian convolution
class GaussianSSE4DKernel;

// Hemi-spherical convolution, for BRDFs 
class HemiSphericalGaussianKernel;
class HemiSphericalSSEGaussianKernel;


#ifndef M_PI
#define M_PI 3.1415926535897932385626
#endif

template<typename T> T sqr(const T x) { return x*x; };

template<typename T>
void matVecMul(const T* M, const T* u, T *v, size_t N) {
#pragma omp parallel for
	for (int i=0; i<N; i++) {
		T r = 0.;
		for (int j=0; j<N; j++) {
			r+= M[i*N+j]*u[j];
		}
		v[i] = r;
	}
}

template<typename T>
void matTransVecMul(const T* M, const T* u, T *v, size_t N) {	
#pragma omp parallel for
	for (int i=0; i<N; i++) {
		T r = 0.;
		for (int j=0; j<N; j++) {
			r+= M[j*N+i]*u[j]; // beurk. one cache miss per access at worst
		}
		v[i] = r;
	}
}
class MatrixKernel {
public:
	MatrixKernel(double* mat = NULL, int nbcols = 0) {
		matrix = mat;
		N = nbcols;
	}

	void convolve(const double* u, double* result) const {
		matVecMul(matrix, u, result, N);
	}
	void convolveAdjoint(const double* u, double* result) const {
		matTransVecMul(matrix, u, result, N);
	}

	double operator()(int i, int j) {
		return matrix[i*N+j];
	}
	int N;
	double* matrix;
};

class SymmetricMatrixKernel {
public:
	SymmetricMatrixKernel(double gamma = 1., double* mat = NULL, int nbcols = 0) {
		matrix = mat;
		N = nbcols;
		this->gamma = gamma;
	}

	void convolve(const double* u, double* result) const {
		matVecMul(matrix, u, result, N);
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double operator()(int i, int j) {
		return matrix[i*N+j];
	}
	int N;
	double* matrix;
	double gamma;
};

class Gaussian1DKernel {
public:
	Gaussian1DKernel(double regularization = 25., int W = 0) {
		gamma = regularization;
		W_ = W;
		N = W_;
		kernel1d = new double[W_];
		for (int i=0; i<W_; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-i*i / gamma));
		}
	}
	Gaussian1DKernel(const Gaussian1DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		N = W_;
		kernel1d = new double[W_];
		std::memcpy(kernel1d, b.kernel1d, W_*sizeof(double));
	}
	Gaussian1DKernel& operator=(const Gaussian1DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		N = W_;
		kernel1d = new double[W_];
		std::memcpy(kernel1d, b.kernel1d, W_*sizeof(double));
		return *this;
	}
	~Gaussian1DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		return std::max(EPSILON, exp(-(sqr(i-j))/gamma));
	}
	void convolve(const double* u, double* result) const {

		double* tmp = new double[W_]; // allocating here ; otherwise not thread-safe

		for (int j=0; j<W_; j++) {
			double conv = 0;
			for (int k=0; k<W_; k++) {
				conv+=kernel1d[abs(j-k)]*u[k];
			}
			tmp[j] = conv;
		}

		std::memcpy(result, tmp, W_*sizeof(double));
		delete[] tmp;
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, N;
	double* kernel1d;
};


class Gaussian2DKernel {
public:
	Gaussian2DKernel(double regularization = 25., int W = 0, int H = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		N = W_*H_;
		kernel1d = new double[std::max(W, H)];
		for (int i=0; i<std::max(W, H); i++) {
			kernel1d[i] = std::max(EPSILON, exp(-i*i / gamma));
		}
	}
	Gaussian2DKernel(const Gaussian2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		kernel1d = new double[std::max(W_, H_)];
		std::memcpy(kernel1d, b.kernel1d, std::max(W_, H_)*sizeof(double));
	}
	Gaussian2DKernel& operator=(const Gaussian2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		kernel1d = new double[std::max(W_, H_)];
		std::memcpy(kernel1d, b.kernel1d, std::max(W_, H_)*sizeof(double));
		return *this;
	}
	~Gaussian2DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		int x0 = i%W_;
		int y0 = i/W_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1))/gamma));
	}
	void convolve(const double* u, double* result) const {

		double* tmp = new double[W_*H_]; // allocating here ; otherwise not thread-safe

		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				double conv = 0;
				for (int k=0; k<W_; k++) {
					conv+=kernel1d[abs(j-k)]*u[i*W_ + k];
				}
				tmp[i+j*H_] = conv;
			}
		}


		for (int j=0; j<W_; j++) {
			for (int i=0; i<H_; i++) {
				double conv = 0;
				for (int k=0; k<H_; k++) {
					conv+=kernel1d[abs(i-k)]*tmp[k + j*H_];
				}
				result[i*W_+j] = conv;
			}
		}
		delete[] tmp;
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_;
	int N;
	double* kernel1d;
};




class TruncatedGaussian2DKernel {
public:
	TruncatedGaussian2DKernel(double regularization = 25., int W = 0, int H = 0, int kernel_size = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		N = W_*H_;
		S = kernel_size;
		if (S==0) S=2*regularization;
		kernel1d = new double[S];
		for (int i=0; i<S; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-i*i / gamma));
		}
	}
	TruncatedGaussian2DKernel(const TruncatedGaussian2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		S = b.S;
		kernel1d = new double[S];
		std::memcpy(kernel1d, b.kernel1d, S*sizeof(double));
	}
	TruncatedGaussian2DKernel& operator=(const TruncatedGaussian2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		S = b.S;
		kernel1d = new double[S];
		std::memcpy(kernel1d, b.kernel1d, S*sizeof(double));
		return *this;
	}
	~TruncatedGaussian2DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		int x0 = i%W_;
		int y0 = i/W_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1))/gamma)); // do not truncate here
	}
	void convolve(const double* u, double* result) const {

		double* tmp = new double[W_*H_]; // allocating here ; otherwise not thread-safe (thread_local not in VS)
		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				double conv = 0;
				for (int l=std::max(-j, -S); l<std::min(W_-j, S); l++) {
					conv+=kernel1d[abs(l)]*u[i*W_ + j+l];
				}
				tmp[i+j*H_] = conv;
			}
		}


		for (int j=0; j<W_; j++) {
			for (int i=0; i<H_; i++) {
				double conv = 0;				
				for (int l=std::max(-i, -S); l<std::min(H_-i, S); l++) {
					conv+=kernel1d[abs(l)]*tmp[i+l + j*H_];
				}
				result[i*W_+j] = conv;
			}
		}
		delete[] tmp;
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_;
	int N, S;
	double* kernel1d;
};


class Gaussian3DKernel {
public:
	Gaussian3DKernel(double regularization = 25., int W = 0, int H = 0, int D = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		D_ = D;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_);
		kernel1d = new double[M];
		for (int i=0; i<M; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-i*i / gamma));
		}
	}
	Gaussian3DKernel(const Gaussian3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_);
		kernel1d = new double[M];
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
	}
	Gaussian3DKernel& operator=(const Gaussian3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_);
		kernel1d = new double[M];
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
		return *this;
	}
	~Gaussian3DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		int z0 = i/(W_*H_);
		i-=z0*W_*H_;
		int x0 = i%W_;
		int y0 = i/W_;

		int z1 = j/(W_*H_);
		j-=z1*W_*H_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1)+sqr(z0-z1))/gamma));
	}
	void convolve(const double* u, double* result) const {

		double* tmp = new double[W_*H_*D_]; // allocating here ; otherwise not thread-safe

		for (int d=0; d<D_; d++) {
			for (int i=0; i<H_; i++) {
				for (int j=0; j<W_; j++) {
					double conv = 0;
					for (int k=0; k<W_; k++) {
						conv+=kernel1d[abs(j-k)]*u[d*W_*H_ + i*W_ + k];
					}
					tmp[d*W_*H_ + j*H_+i] = conv;
				}
			}


			for (int j=0; j<W_; j++) {
				for (int i=0; i<H_; i++) {
					double conv = 0;
					for (int k=0; k<H_; k++) {
						conv+=kernel1d[abs(i-k)]*tmp[d*W_*H_ + j*H_ + k];
					}
					result[(i*W_+j)*D_ + d] = conv;
				}
			}
		}

		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				for (int d=0; d<D_; d++) {
					double conv = 0;
					for (int k=0; k<D_; k++) {
						conv+=kernel1d[abs(d-k)]*result[k + (i*W_ + j)*D_];
					}
					tmp[d*W_*H_ + i*W_+j] = conv;
				}
			}
		}
		std::memcpy(result, tmp, W_*H_*D_*sizeof(double));
		delete[] tmp;
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_, D_, N;
	double* kernel1d;
};






class TruncatedGaussian3DKernel {
public:
	TruncatedGaussian3DKernel(double regularization = 25., int W = 0, int H = 0, int D = 0, int kernel_size = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		D_ = D;
		N = W_*H_*D_;		
		S = kernel_size;
		if (kernel_size==0) kernel_size = 2*regularization;
		kernel1d = new double[S];
		for (int i=0; i<S; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-i*i / gamma));
		}
	}
	TruncatedGaussian3DKernel(const TruncatedGaussian3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		N = W_*H_*D_;
		S = b.S;		
		kernel1d = new double[S];
		std::memcpy(kernel1d, b.kernel1d, S*sizeof(double));
	}
	TruncatedGaussian3DKernel& operator=(const TruncatedGaussian3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		S = b.S;
		N = W_*H_*D_;		
		kernel1d = new double[S];
		std::memcpy(kernel1d, b.kernel1d, S*sizeof(double));
		return *this;
	}
	~TruncatedGaussian3DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		int z0 = i/(W_*H_);
		i-=z0*W_*H_;
		int x0 = i%W_;
		int y0 = i/W_;

		int z1 = j/(W_*H_);
		j-=z1*W_*H_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1)+sqr(z0-z1))/gamma)); // not truncated here
	}
	void convolve(const double* u, double* result) const {

		double* tmp = new double[W_*H_*D_]; // allocating here ; otherwise not thread-safe

#pragma omp parallel for
		for (int d=0; d<D_; d++) {
			for (int i=0; i<H_; i++) {
				for (int j=0; j<W_; j++) {
					double conv = 0;
					for (int l=std::max(-j,-S); l<std::min(W_-j, S); l++) {
						conv+=kernel1d[abs(l)]*u[d*W_*H_ + i*W_ + j+l];
					}
					tmp[d*W_*H_ + j*H_+i] = conv;
				}
			}


			for (int j=0; j<W_; j++) {
				for (int i=0; i<H_; i++) {
					double conv = 0;
					for (int l=std::max(-i, -S); l<std::min(H_-i, S); l++) {
						conv+=kernel1d[abs(l)]*tmp[d*W_*H_ + j*H_ + i+l];
					}
					result[(i*W_+j)*D_ + d] = conv;
				}
			}
		}
#pragma omp parallel for
		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				for (int d=0; d<D_; d++) {
					double conv = 0;
					for (int l=std::max(-d, -S); l<std::min(D_-d, S); l++) {
						conv+=kernel1d[l]*result[d+l + (i*W_ + j)*D_];
					}
					tmp[d*W_*H_ + i*W_+j] = conv;
				}
			}
		}
		std::memcpy(result, tmp, W_*H_*D_*sizeof(double));
		delete[] tmp;
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_, D_, N, S;
	double* kernel1d;
};




class GaussianSSE2DKernel {
public:
	GaussianSSE2DKernel(double regularization = 25., int W = 0, int H = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		N = W_*H_;
		C_ = std::max(W, H);
		kernel1d = (double*)malloc_simd(std::max(W, H)*2*sizeof(double), ALIGN);
		for (int i=0; i<std::max(W, H)*2; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-sqr(C_-i) / gamma));
		}
		int nbthreads = omp_get_max_threads();
		tmpmem = (double*)malloc_simd(W_*H_*nbthreads*sizeof(double), ALIGN);
	}
	GaussianSSE2DKernel(const GaussianSSE2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		C_ = std::max(W_, H_);
		int nbthreads = omp_get_max_threads();
		kernel1d = (double*)malloc_simd(std::max(W_, H_)*2*sizeof(double), ALIGN);
		std::memcpy(kernel1d, b.kernel1d, std::max(W_, H_)*2*sizeof(double));
		tmpmem = (double*)malloc_simd(W_*H_*nbthreads*sizeof(double), ALIGN);
	}
	GaussianSSE2DKernel& operator=(const GaussianSSE2DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		N = W_*H_;
		C_ = std::max(W_, H_);
		int nbthreads = omp_get_max_threads();
		kernel1d = (double*)malloc_simd(std::max(W_, H_)*2*sizeof(double), ALIGN);
		std::memcpy(kernel1d, b.kernel1d, std::max(W_, H_)*2*sizeof(double));
		tmpmem = (double*)malloc_simd(W_*H_*nbthreads*sizeof(double), ALIGN);
		return *this;
	}
	~GaussianSSE2DKernel() {
		free_simd(kernel1d);
		free_simd(tmpmem);
	}
	double operator()(int i, int j) {
		int x0 = i%W_;
		int y0 = i/W_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1))/gamma));
	}
	void convolve(const double* u, double* result) const {

		double* tmp = tmpmem + omp_get_thread_num() * W_*H_;  // convolve might be called in parallel	

#pragma omp parallel for
		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				tmp[i+j*H_] = dotp_full(&kernel1d[C_-j], &u[i*W_], W_);
			}
		}

#pragma omp parallel for
		for (int j=0; j<W_; j++) {
			for (int i=0; i<H_; i++) {
				result[i*W_+j] = dotp_full(&kernel1d[C_-i], &tmp[j*H_], H_);
			}
		}
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_, C_;
	int N;
	double* kernel1d;
	double* tmpmem;
};


class GaussianSSE3DKernel {
public:
	GaussianSSE3DKernel(double regularization = 25., int W = 0, int H = 0, int D = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		D_ = D;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_)*2;
		C_ = M/2;
		kernel1d = (double*)malloc_simd(M*sizeof(double), ALIGN);
		for (int i=0; i<M; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-sqr(i-C_) / gamma));
		}
		int nbthreads = omp_get_max_threads();
		tmpmem = (double*)malloc_simd(W_*H_*D_*nbthreads*sizeof(double), ALIGN);
	}
	GaussianSSE3DKernel(const GaussianSSE3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_)*2;
		C_ = M/2;
		kernel1d = (double*)malloc_simd(M*sizeof(double), ALIGN);
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
		int nbthreads = omp_get_max_threads();
		tmpmem = (double*)malloc_simd(W_*H_*D_*nbthreads*sizeof(double), ALIGN);
	}
	GaussianSSE3DKernel& operator=(const GaussianSSE3DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		N = W_*H_*D_;
		int M = std::max(std::max(W_, H_), D_)*2;
		C_ = M/2;
		kernel1d = (double*)malloc_simd(M*sizeof(double), ALIGN);
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
		int nbthreads = omp_get_max_threads();
		tmpmem = (double*)malloc_simd(W_*H_*D_*nbthreads*sizeof(double), ALIGN);

		return *this;
	}
	~GaussianSSE3DKernel() {
		free_simd(tmpmem);
		free_simd(kernel1d);
	}
	double operator()(int i, int j) {
		int z0 = i/(W_*H_);
		i-=z0*W_*H_;
		int x0 = i%W_;
		int y0 = i/W_;

		int z1 = j/(W_*H_);
		j-=z1*W_*H_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1)+sqr(z0-z1))/gamma));
	}
	void convolve(const double* u, double* result) const {
		
		double* tmp = tmpmem+omp_get_thread_num() * W_*H_*D_;  // convolve might be called in parallel
#pragma omp parallel for
		for (int d=0; d<D_; d++) {
			for (int i=0; i<H_; i++) {
				const double* ud = &u[d*W_*H_ + i*W_];
				for (int j=0; j<W_; j++) {
					tmp[d*W_*H_ + j*H_+i] = dotp_full(&kernel1d[C_-j], ud, W_);
				}
			}


			for (int j=0; j<W_; j++) {
				const double* ud = &tmp[d*W_*H_ + j*H_];
				for (int i=0; i<H_; i++) {
					result[(i*W_+j)*D_ + d] = dotp_full(&kernel1d[C_-i], ud, H_);
				}
			}
		}

#pragma omp parallel for
		for (int i=0; i<H_; i++) {
			for (int j=0; j<W_; j++) {
				const double* ud = &result[(i*W_ + j)*D_];
				for (int d=0; d<D_; d++) {
					tmp[d*W_*H_ + i*W_+j] = dotp_full(&kernel1d[C_-d], ud, D_);
				}
			}
		}
		std::memcpy(result, tmp, W_*H_*D_*sizeof(double));		
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_, D_, N, C_;
	double* kernel1d;
	double* tmpmem;
};




class GaussianSSE4DKernel {
public:
	GaussianSSE4DKernel(double regularization = 25., int W = 0, int H = 0, int D = 0, int T = 0) {
		gamma = regularization;
		W_ = W;
		H_ = H;
		D_ = D;
		T_ = T;
		N = W_*H_*D_*T_;
		int M = std::max(std::max(W_, H_), std::max(T_, D_))*2;
		C_ = M/2;
		kernel1d = new double[M];
		for (int i=0; i<M; i++) {
			kernel1d[i] = std::max(EPSILON, exp(-sqr(i-C_) / gamma));
		}
	}
	GaussianSSE4DKernel(const GaussianSSE4DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		T_ = b.T_;
		N = W_*H_*D_*T_;
		int M = std::max(std::max(W_, H_), std::max(T_, D_))*2;
		C_ = M/2;
		kernel1d = new double[M];
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
	}
	GaussianSSE4DKernel& operator=(const GaussianSSE4DKernel& b) {
		gamma = b.gamma;
		W_ = b.W_;
		H_ = b.H_;
		D_ = b.D_;
		T_ = b.T_;
		N = W_*H_*D_*T_;
		int M = std::max(std::max(W_, H_), std::max(T_, D_))*2;
		C_ = M/2;
		kernel1d = new double[M];
		std::memcpy(kernel1d, b.kernel1d, M*sizeof(double));
		return *this;
	}
	~GaussianSSE4DKernel() {
		delete[] kernel1d;
	}
	double operator()(int i, int j) {
		int w0 = i/(W_*H_*D_);
		i-=w0*W_*H_*D_;
		int z0 = i/(W_*H_);
		i-=z0*W_*H_;
		int x0 = i%W_;
		int y0 = i/W_;

		int w1 = j/(W_*H_*D_);
		j-=w1*W_*H_*D_;
		int z1 = j/(W_*H_);
		j-=z1*W_*H_;
		int x1 = j%W_;
		int y1 = j/W_;
		return std::max(EPSILON, exp(-(sqr(x0-x1)+sqr(y0-y1)+sqr(z0-z1)+sqr(w0-w1))/gamma));
	}
	void convolve(const double* u, double* result) const {

		double* tmp = (double*)malloc_simd(W_*H_*D_*T_*sizeof(double), ALIGN); // allocating here ; otherwise not thread-safe		
		double* tmp2 = (double*)malloc_simd(W_*H_*D_*T_*sizeof(double), ALIGN); // allocating here ; otherwise not thread-safe	
		
#pragma omp parallel for
		for (int t=0; t<T_; t++) {
			for (int d=0; d<D_; d++) {
				for (int i=0; i<H_; i++) {
					const double* ud = &u[t*W_*H_*D_ + d*W_*H_ + i*W_];
					for (int j=0; j<W_; j++) {
						tmp[t*W_*H_*D_ + d*W_*H_ + j*H_+i] = dotp_full(&kernel1d[C_-j], ud, W_);
					}
				}


				for (int j=0; j<W_; j++) {
					const double* ud = &tmp[t*W_*H_*D_ + d*W_*H_ + j*H_];
					for (int i=0; i<H_; i++) {
						result[t*W_*H_*D_ + (i*W_+j)*D_ + d] = dotp_full(&kernel1d[C_-i], ud, H_);
					}
				}
			}

			for (int i=0; i<H_; i++) {
				for (int j=0; j<W_; j++) {
					const double* ud = &result[t*W_*H_*D_ + (i*W_ + j)*D_];
					for (int d=0; d<D_; d++) {
						tmp2[(d*W_*H_ + i*W_+j)*T_ + t] = dotp_full(&kernel1d[C_-d], ud, D_);
					}
				}
			}
		}
		
#pragma omp parallel for
		for (int d=0; d<D_; d++) {
			for (int i=0; i<H_; i++) {
				for (int j=0; j<W_; j++) {
					const double* ud = &tmp2[(d*W_*H_ + i*W_+j)*T_];
					for (int t=0; t<T_; t++) {
						result[t*W_*H_*D_ + d*W_*H_ + i*W_+j] = dotp_full(&kernel1d[T_-t], ud, T_);
					}
				}
			}
		}
		free_simd(tmp);
		free_simd(tmp2);
	}
	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}

	double gamma;
	int W_, H_, D_, T_, N, C_;
	double* kernel1d;
};


class HemiSphericalGaussianKernel {
public:
	HemiSphericalGaussianKernel(double regularization = 25., int NTheta = 0, int NPhi = 0) {
		NTheta_ = NTheta;
		NPhi_ = NPhi;
		gamma = regularization;
		N = NTheta_*NPhi_;
		sinTheta.resize(NTheta_);
		sinPhi.resize(NPhi_);
		cosTheta.resize(NTheta_);
		cosPhi.resize(NPhi_);
		for (int i=0; i<NTheta_; i++) {
			sinTheta[i] = std::sin((i+0.1)*M_PI*0.5/NTheta);
			cosTheta[i] = std::cos((i+0.1)*M_PI*0.5/NTheta);
		}
		for (int i=0; i<NPhi_; i++) {
			sinPhi[i] = std::sin((i+0.01)*M_PI*2./NPhi_);
			cosPhi[i] = std::cos((i+0.01)*M_PI*2./NPhi_);
		}
	}
	std::vector<double> sinTheta, sinPhi, cosTheta, cosPhi;

	inline double dist_sphere(int itheta0, int iphi0, int itheta1, int iphi1) const {

		double s0 = sinTheta[itheta0];
		double c0 = cosTheta[itheta0];
		double s1 = sinTheta[itheta1];
		double c1 = cosTheta[itheta1];
		double cd = cosPhi[iphi0]*cosPhi[iphi1]+sinPhi[iphi0]*sinPhi[iphi1];
		double d = std::acos(std::max(1E-12, std::min(1.-1E-12, s0*s1*cd + c0*c1)));
		return d;
	}

	double operator()(int i, int j) {
		int x0 = i%NPhi_;
		int y0 = i/NPhi_;
		int x1 = j%NPhi_;
		int y1 = j/NPhi_;
		double d = dist_sphere(x0, y0, x1, y1);
		return std::max(EPSILON, exp(-(d*d)/gamma));
	}

	void convolve(const double* u, double* result) const {


#pragma omp parallel for
		for (int i=0; i<NTheta_; i++) {
			double s0 = sinTheta[i];
			double c0 = cosTheta[i];
			for (int j=0; j<NPhi_; j++) {
				double r = 0;
				double cp0 = cosPhi[j];
				double sp0 = sinPhi[j];
				for (int k=0; k<NTheta_; k++) {
					double s1 = sinTheta[k];
					double c1 = cosTheta[k];
					double c0c1 = c0*c1;
					double s0s1 = s0*s1;
					for (int l=0; l<NPhi_; l++) {																		
						double cd = cp0*cosPhi[l]+sp0*sinPhi[l];
						double d = std::acos(std::max(1E-12, std::min(1.-1E-12, s0s1*cd + c0c1)));						
						r+=u[k*NPhi_+l]*std::max(EPSILON, exp(-(d*d)/gamma)) *s1;
					}
				}
				result[i*NPhi_ + j] = r;
			}
		}

	}

	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}


	int NTheta_, NPhi_, N;
	double gamma;
};


class HemiSphericalSSEGaussianKernel {
public:
	HemiSphericalSSEGaussianKernel(double regularization = 25., int NTheta = 0, int NPhi = 0) {
		NTheta_ = NTheta;
		NPhi_ = NPhi;
		gamma = regularization;
		N = NTheta_*NPhi_;
		sinTheta.resize(NTheta_);		
		cosTheta.resize(NTheta_);

		sinPhi = (double*) malloc_simd(NPhi_*sizeof(double), ALIGN);
		cosPhi = (double*) malloc_simd(NPhi_*sizeof(double), ALIGN);
		for (int i=0; i<NTheta_; i++) {
			sinTheta[i] = std::sin((i+0.1)*M_PI*0.5/NTheta);
			cosTheta[i] = std::cos((i+0.1)*M_PI*0.5/NTheta);
		}
		for (int i=0; i<NPhi_; i++) {
			sinPhi[i] = std::sin((i+0.01)*M_PI*2./NPhi_);
			cosPhi[i] = std::cos((i+0.01)*M_PI*2./NPhi_);
		}
	}
	HemiSphericalSSEGaussianKernel(const HemiSphericalSSEGaussianKernel& b) {
		gamma = b.gamma;
		NTheta_ = b.NTheta_;
		NPhi_ = b.NPhi_;
		N = NTheta_*NPhi_;
		sinTheta = b.sinTheta;
		cosTheta = b.cosTheta;
		sinPhi = (double*)malloc_simd(NPhi_*sizeof(double), ALIGN);
		cosPhi = (double*)malloc_simd(NPhi_*sizeof(double), ALIGN);
		std::memcpy(sinPhi, b.sinPhi, NPhi_*sizeof(double));
		std::memcpy(cosPhi, b.cosPhi, NPhi_*sizeof(double));
	}
	HemiSphericalSSEGaussianKernel& operator=(const HemiSphericalSSEGaussianKernel& b) {
		gamma = b.gamma;
		NTheta_ = b.NTheta_;
		NPhi_ = b.NPhi_;
		N = NTheta_*NPhi_;
		sinTheta = b.sinTheta;
		cosTheta = b.cosTheta;

		sinPhi = (double*)malloc_simd(NPhi_*sizeof(double), ALIGN);
		cosPhi = (double*)malloc_simd(NPhi_*sizeof(double), ALIGN);
		std::memcpy(sinPhi, b.sinPhi, NPhi_*sizeof(double));
		std::memcpy(cosPhi, b.cosPhi, NPhi_*sizeof(double));
		return *this;
	}
	std::vector<double> sinTheta, cosTheta;
	double *sinPhi, *cosPhi;

	~HemiSphericalSSEGaussianKernel() {
		free_simd(sinPhi);
		free_simd(cosPhi);
	}
	double operator()(int i, int j) {
		/*int x0 = i%NPhi_;
		int y0 = i/NPhi_;
		int x1 = j%NPhi_;
		int y1 = j/NPhi_;
		double d = dist_sphere(x0, y0, x1, y1);
		return std::max(EPSILON, exp(-(d*d)/gamma));*/
		assert(false);
		return 0;
	}

	void convolve(const double* u, double* result) const {

		__m256d zero = _mm256_set1_pd(1E-12);
		__m256d one = _mm256_set1_pd(1.-1E-12);		
		__m256d epsi = _mm256_set1_pd(EPSILON);
		__m256d mgamma = _mm256_set1_pd(-gamma);

#pragma omp parallel for
		for (int i=0; i<NTheta_; i++) {
			double st0 = sinTheta[i];
			double ct0 = cosTheta[i];

			for (int j=0; j<NPhi_; j++) {
				__m256d rd = zero;
				double cp1 = cosPhi[j];
				double sp1 = sinPhi[j];
				__m256d sp1d = _mm256_set1_pd(sp1);
				__m256d cp1d = _mm256_set1_pd(cp1);
				for (int k=0; k<NTheta_; k++) {
					double st1 = sinTheta[k];
					double ct1 = cosTheta[k];
					__m256d st1d = _mm256_set1_pd(st1);
					__m256d ct0ct1 = _mm256_set1_pd(ct0*ct1);
					__m256d st0st1 = _mm256_set1_pd(st0*st1);
					for (int l=0; l<NPhi_; l+=4) {  // ASSUMES NPHI%4 == 0 ; this is ok for our BRDFs						
						__m256d ud = simd_load(u + k*NPhi_+l);
						__m256d cd = _mm256_add_pd(_mm256_mul_pd(cp1d, simd_load(cosPhi+l)), _mm256_mul_pd(sp1d, simd_load(sinPhi+l)));
						__m256d angle = _mm256_acos_pd(_mm256_max_pd(zero, _mm256_min_pd(one, _mm256_add_pd(_mm256_mul_pd(st0st1, cd), ct0ct1))));
						__m256d conv = _mm256_mul_pd(_mm256_max_pd(epsi, _mm256_exp_pd(_mm256_div_pd(_mm256_mul_pd(angle, angle), mgamma))), st1d);
						__m256d addi = _mm256_mul_pd(ud, conv);
						rd = _mm256_add_pd(rd, addi);	
					}
				}
				
				m256d rrd; rrd.variable = rd;
				result[i*NPhi_ + j] = rrd.m256d_f64[0] + rrd.m256d_f64[1] + rrd.m256d_f64[2]+rrd.m256d_f64[3];
			}
		}

	}

	void convolveAdjoint(const double* u, double* result) const {
		convolve(u, result);
	}


	int NTheta_, NPhi_, N;
	double gamma;
};