// Author: Nicolas Bonneel
// Released: April 14th, 2016
// Credit the following article: 
//   @article{BPC16, 
//   author = {Bonneel, Nicolas and Peyr\'{e}, Gabriel and Cuturi, Marco}, 
//   title = {{Wasserstein Barycentric Coordinates: Histogram Regression Using Optimal Transport}}, 
//   journal = {ACM Transactions on Graphics (SIGGRAPH 2016)}, 
//   volume = {35}, 
//   number = {4}, 
//   year = {2016}, 
//   }
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#pragma once

#include "sse_helpers.h"

template<typename T> T safe_log(const T x) {
	return std::log(std::max(EPSILON, x));
};

template<class Kernel>
class QuadraticLoss {
public:
	QuadraticLoss() {};
	QuadraticLoss(Kernel &ker) {};

	double loss(const double* v1, const double* v2, int N) const {
		double r = 0;
		for (int i=0; i<N; i++) {
			double d = v1[i]-v2[i];
			r += d*d;
		}
		return r;
	}

	void gradient(const double* v1, const double* v2, int N, double* res) const {

		for (int i=0; i<N; i++) {
			res[i] = 2.*(v1[i] - v2[i]);
		}

	}
};

template<class Kernel>
class TVLoss {
public:
	TVLoss() {};
	TVLoss(Kernel &ker) {};

	double loss(const double* v1, const double* v2, int N) const {
		double r = 0;
		for (int i=0; i<N; i++) {
			double d = v1[i]-v2[i];
			r += abs(d);
		}
		return 0.5*r;
	}

	void gradient(const double* v1, const double* v2, int N, double* res) const {

		for (int i=0; i<N; i++) {
			double d = v1[i]-v2[i];
			res[i] = (d>0?0.5:-0.5);
		}

	}
};

template<class Kernel>
class KLLoss {
public:
	KLLoss() {};
	KLLoss(Kernel &ker) {};

	double loss(const double* v1, const double* v2, int N) const {
		double r = 0;
		for (int i=0; i<N; i++) {
			double d = v1[i]*safe_log(v1[i]/v2[i]);
			r += d;
		}
		return r;
	}

	void gradient(const double* v1, const double* v2, int N, double* res) const {

		for (int i=0; i<N; i++) {
			res[i] = 1. + safe_log(v1[i] / v2[i]);
		}

	}
};

template<class Kernel>
class WassersteinLoss {
public:
	WassersteinLoss() {};
	WassersteinLoss(Kernel &ker) { kernel = ker; };


	Kernel kernel;

	double loss(const double* v1, const double* v2, int N) const {
		
		int Niters = 50;
		std::vector<double, aligned_allocator<double> > a(N, 1.), b(N, 1.), convolution(N);		

		// Bregman Projections
		for (int iter=0; iter<Niters; iter++) {
			
			kernel.convolveAdjoint(&b[0], &convolution[0]);
			for (int j=0; j<N; j++) {
				a[j] = v1[j] / convolution[j];
			}			
			kernel.convolve(&a[0], &convolution[0]);
			for (int j=0; j<N; j++) {
				b[j] = v2[j] / convolution[j];
			}	
		}

		double l = 0;
		kernel.convolveAdjoint(&b[0], &convolution[0]);
		for (int j=0; j<N; j++) {
			//l+= v1[j]*log(a[j]) + v2[j]*log(b[j]);
			l += safe_log(a[j])*v1[j] + safe_log(b[j])*v2[j] -  a[j]*convolution[j];
		}
		return kernel.gamma*l;
	}

	void gradient(const double* v1, const double* v2, int N, double* res) const {

		int Niters = 50;
		std::vector<double, aligned_allocator<double> > a(N, 1.), b(N, 1.), convolution(N);		

		// Bregman Projections
		for (int iter=0; iter<Niters; iter++) {

			kernel.convolveAdjoint(&b[0], &convolution[0]);
			for (int j=0; j<N; j++) {
				a[j] = v1[j] / convolution[j];
			}
			kernel.convolve(&a[0], &convolution[0]);
			for (int j=0; j<N; j++) {
				b[j] = v2[j] / convolution[j];
			}
		}
		
		for (int i=0; i<N; i++) {
			res[i]= kernel.gamma*safe_log(a[i]);
		}		
	}

};


static double* mask; // ugly -- I don't want to change my entire code to pass the mask in the MaskedTVLoss constructor (and hence, the others too)

template<class Kernel>
class MaskedTVLoss {
public:
	MaskedTVLoss() {};
	MaskedTVLoss(Kernel &ker) {		
	};
	double loss(const double* v1, const double* v2, int N) const {
		double r = 0;
		for (int i=0; i<N; i++) {
			double d = (v1[i]-v2[i])*mask[i];
			r += abs(d);
		}
		return 0.5*r;
	}

	void gradient(const double* v1, const double* v2, int N, double* res) const {

		for (int i=0; i<N; i++) {
			double d = v1[i]-v2[i];
			res[i] = (d>0?0.5:-0.5) *mask[i];
		}
	}	
};