#include "inverseWasserstein.h"
#include <sstream>
#include <omp.h>

// test unit for barycentric coordinates computation

void generate_gaussian(double mx, double my, double sx, double sy, int N, std::vector<double> &gauss) {
	gauss.resize(N*N);
	double s = 0;
	for (int i=0; i<N; i++) {
		for (int j=0; j<N; j++) {
			gauss[i*N+j] = exp( (i-mx)*(i-mx)/(2*sx*sx) + (j-my)*(j-my)/(2*sy*sy) );
			s+= gauss[i*N+j];
		}
	}
	for (int i=0; i<N*N; i++) {
		gauss[i] /= s;
	}
}


int main() {
	
	int S = 5;          // number of histograms in the database
	std::vector<std::vector<double> > pdfs(S);  // histogram database
	std::vector<double> target_pdf;   // histogram to retrieve barycentric coordinates 
	int nbins = 128;	// discretization. Let's use 128x128 2-d histograms
	double gamma = 50;   // regularization parameter
	
	generate_gaussian(30, 30, 30., 60., nbins, pdfs[0]);
	generate_gaussian(80, 100, 25., 30., nbins, pdfs[1]);
	generate_gaussian(60, 70, 45, 25., nbins, pdfs[2]);
	generate_gaussian(95, 35, 25., 35., nbins, pdfs[3]);
	generate_gaussian(50, 90, 30., 20., nbins, pdfs[4]);	

	// we synthesize an histogram to fit by computing the forward Wasserstein barycenter problem
	double ground_truth_weights[5] = {0.1, 0.2, 0.3, 0.35, 0.05}; 
	GaussianSSE2DKernel gibbs_kernel(gamma, nbins, nbins);
	Problem<GaussianSSE2DKernel> forward_problem(pdfs, std::vector<double>(), gibbs_kernel, ground_truth_weights, 1);
	WassersteinBarycenter<GaussianSSE2DKernel> barycenter(&forward_problem, 100);
	barycenter.compute_barycenter();
	target_pdf = barycenter.get_barycenter();
	
	
	// we now try to recover the ground truth weights automatically via our Wasserstein Barycentric Coordinates. 	

	std::vector<double> nearest_pdf;
	std::vector<double> lambdas(S, 1./S);	
	regress<GaussianSSE2DKernel, QuadraticLoss<GaussianSSE2DKernel> >(pdfs, target_pdf, gibbs_kernel, nearest_pdf, &lambdas[0], 100);
	
	std::cout<<"ground truth / result: "<<std::endl;
	for (int i=0; i<S; i++) {
		std::cout<< ground_truth_weights[i] << " / " << lambdas[i] <<std::endl;
	}
	
	return 0;
}

